unit Main;
{ Recovers DCR files trashed by Image Editor
  Freeware for Delphi 1.X (untested on Delphi 2.0)
  Release 1.0, March 1996 by Grant Walker
  Comments welcome - gw@anfi.pacit.tas.gov.au
}
interface
{$A-}  { Streaming fails if word alignment enabled }
uses
  Classes, Forms, Dialogs, StdCtrls, Controls,
  Buttons, SysUtils, IniFiles;

type
  TMainForm = class(TForm)
    SrcEdit: TEdit;
    DstEdit: TEdit;
    BrowseSrcButton: TButton;
    FirstAidButton: TBitBtn;
    OpenFileDlg: TOpenDialog;
    CommentMemo: TMemo;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    ShowPalette: TCheckBox;
    ShowRasters: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure CloseButtonClick(Sender: TObject);
    procedure BrowseSrcButtonClick(Sender: TObject);
    procedure FirstAidButtonClick(Sender: TObject);
    procedure EditKeyUp(Sender:TObject; var Key:Word; Shift:TShiftState);
    procedure SrcEditChange(Sender: TObject);
    procedure CheckEscapeKeyPress(Sender: TObject; var Key: Char);
  private
    ini: TIniFile;
    Cancel: boolean;
    procedure AttemptRecovery;
    procedure Comment(s:string);
    procedure AppException(Sender: TObject; E: Exception);
  end;

var
  MainForm: TMainForm;

implementation
uses Messages,WinProcs,WinTypes;
{$R *.DFM}

procedure TMainForm.FormCreate(Sender: TObject);
const
  MemoTabs:array[1..16] of integer = (21,32,43,54,65,76,87,98,109,120,131,142,153,164,175,186);
begin
  Application.OnException := AppException;
  ini := TIniFile.Create('win.ini');
  OpenFileDlg.InitialDir := ini.ReadString('dcrfix','dir','');
  SendMessage(CommentMemo.Handle,EM_SetTabStops,16,Longint(@MemoTabs));
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  ini.Free;
end;

procedure TMainForm.CloseButtonClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.AppException(Sender: TObject; E: Exception);
begin
  Comment(E.Message);
end;

procedure TMainForm.Comment(s:string);
begin
  CommentMemo.Lines.Add(s);
end;

procedure TMainForm.BrowseSrcButtonClick(Sender: TObject);
{ Known bug if SrcEdit = DstEdit, but not worth fixing }
var DLen:integer;
    NewDir:string;
begin
  OpenFileDlg.FileName := SrcEdit.Text;
  with OpenFileDlg do
   if Execute then
     begin
       SrcEdit.Text := FileName;
       NewDir := ExtractFilePath(FileName);
       DLen := length(NewDir);
       if (NewDir[DLen] = '\') and (DLen > 3) then Delete(NewDir,DLen,1);
       ChDir(NewDir);
       ini.WriteString('dcrfix','dir',NewDir);
       FirstAidButton.SetFocus;
     end;
end;

procedure TMainForm.SrcEditChange(Sender: TObject);
begin
  CommentMemo.Lines.Clear;
  FirstAidButton.Enabled := length(SrcEdit.Text) > 0;
end;

procedure TMainForm.FirstAidButtonClick(Sender: TObject);
begin
  AttemptRecovery;
end;

procedure TMainForm.EditKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (Key=13) and FirstAidButton.Enabled then
    begin
      AttemptRecovery;
      Key := 0;
    end;
end;

procedure TMainForm.AttemptRecovery;
type
  EFatalDCR = Exception;
var
  fsrc,fdst:TFileStream;
  i,j,len,palcolours,remain,rasterpos: integer;
  DcrHeader: record
              a,b,c:Byte;
              name:array[0..99] of char;
             end;
  DcrInfo:   record
              Magic: Word;
              BmpSize: LongInt;
             end;
  bmihdr:TBitmapInfoHeader;
  rgb:TRGBQuad;
  Pixel:byte;
  s:string;
begin
  Cancel := False;
  CommentMemo.Lines.Clear;
  fsrc := TFileStream.Create(SrcEdit.Text,fmOpenRead or fmShareDenyWrite);
  Comment(Format('Physical file size is %d',[fsrc.Size]));
  fdst := TFileStream.Create(DstEdit.Text,fmCreate);
  Screen.Cursor := crHourGlass;
  try
      fsrc.Read(DcrHeader,sizeof(DcrHeader));
      with DcrHeader do
        begin
          if (a <> $FF) or (b <> $2) or (c <> $0) then raise EFatalDCR.Create('Header bad: is this a DCR?');
          len := strlen(name);
          if (len = 0) or (len > 64) then raise EFatalDCR.Create('DCR name is unrecoverable');
          Comment(Format('Name is "%s"',[StrPas(name)]));
          fdst.Write(DcrHeader,3+len+1);  { abc & name & null terminator }
        end;
      fsrc.Seek(fdst.Position,soFromBeginning);
      fsrc.Read(DcrInfo,sizeof(DcrInfo));
      fdst.Write(DcrInfo,sizeof(DcrInfo));
      {Comment(Format('Magic word = $%.4x',[DcrInfo.Magic]));}
      with DcrInfo do
        if fdst.Position+BmpSize <> fsrc.Size
          then Comment(Format('File size mismatch!  Hdr(%d) + Bmp(%d) = %d',
                              [fdst.Position,BmpSize,fdst.Position+BmpSize]));
      fsrc.Read(bmihdr,sizeof(bmihdr));
      with bmihdr do
       begin
         if biSize <> sizeof(TBitmapInfoHeader) then raise EFatalDCR.Create('Bad TBITMAPINFOHEADER biSize: '+IntToStr(biSize));
         if (biWidth <> 56) or (biHeight <> 28) then
           Comment(Format('Warning: bitmap is %d x %d',[biWidth,biHeight]));
         if biPlanes <> 1 then raise EFatalDCR.Create('Bad biPlanes');
         case biBitCount of
           1: palcolours := 2;
           4: palcolours := 16;
           8: palcolours := 256;
           { 24 bit colour not supported }
           else raise EFatalDCR.Create('Bad biBitCount');
         end;
         case biCompression of
           bi_RGB: Comment(Format('Bitmap is %d colour, uncompressed',[palcolours]));
           bi_RLE8: if palcolours <> 256 then raise EFatalDCR.Create('RLE8 mismatch')
                    else Comment('Bitmap RLE8 compressed for 256 colours');
           bi_RLE4: if palcolours <> 16 then raise EFatalDCR.Create('RLE4 mismatch')
                    else Comment('Bitmap RLE4 compressed for 16 colours');
           3: Comment(Format('%d colour, compressed using BI_BITFIELDS',[palcolours]));
           else raise EFatalDCR.Create('Bad biCompression');
         end;
         { ignore biXPelsPerMeter..biClrUsed }
       end;
      fdst.Write(bmihdr,sizeof(bmihdr));
      if not ShowPalette.Checked then
        fdst.CopyFrom(fsrc,palcolours*sizeof(TRGBQuad))
      else
        begin
          Comment(#13#10'biClrImportant = '+IntToStr(bmihdr.biClrImportant));
          for i := 1 to palcolours do
            begin
              if Cancel then raise EFatalDCR.Create('Cancelled');
              fsrc.Read(rgb,sizeof(TRGBQuad));
              with rgb do if rgbReserved <> 0 then
                Comment(Format('Palette colour #%d is corrupt (%d)',[i-1,rgbReserved]))
              else if ShowPalette.Checked then
                Comment(Format('Palette colour #%d is (%d,%d,%d)',[i-1,rgbRed,rgbGreen,rgbBlue]));
              fdst.Write(rgb,sizeof(TRGBQuad));
              Application.ProcessMessages;
            end;
        end;
      Comment(#13#10+Format('Rasters should occupy %d bytes',[bmihdr.biSizeImage]));
      rasterpos := fsrc.Position;
      remain := fsrc.Size - rasterpos;
      with bmihdr do
        if remain < biSizeImage then
          begin
            Comment(Format('Only %d bytes remain, padding...',[remain]));
            fdst.CopyFrom(fsrc,remain);
            Pixel := $00;
            for i := 1 to biSizeImage-remain do
              fdst.Write(Pixel,1);
          end
        else
          begin
            if remain = biSizeImage
              then Comment('Copying...')
              else Comment(Format('%d bytes remain, truncating to %d...',[remain,biSizeImage]));
            fdst.CopyFrom(fsrc,biSizeImage);
          end;
    if ShowRasters.Checked then
      begin
        Comment('');
        fsrc.Seek(rasterpos,soFromBeginning);
        for i := 0 to pred(remain) div 16 do
          begin
            if Cancel then raise EFatalDCR.Create('Cancelled');
            s := Format('%.4x:',[i*16]);
            j := 0;
            while (j < 16) and (i*16+j < remain) do
              begin
                fsrc.Read(Pixel,1);
                AppendStr(s,#9+IntToHex(Pixel,2));
                inc(j);
              end;
            Comment(s);
            Application.ProcessMessages;
          end;
        Comment(IntToStr(remain)+' raster bytes in '+SrcEdit.Text+#13#10);
      end;
    Comment('Done.');
  finally
    fdst.Free;
    fsrc.Free;
    Screen.Cursor := crDefault;
    CommentMemo.SetFocus;
  end;
end;

procedure TMainForm.CheckEscapeKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #27 then Cancel := True;
end;

end.
